#include <Servo.h>

#define DEBUG

#include "Debug.h"
#include "Encoder.h"
#include "Ping.h"
#include "Motor.h"

/*
** -------------------------------  Constants and Pin Definitions --------------------------
*/

#define PIN_LED         13

#define PIN_PING         9
#define PIN_PING_SERVO   8
#define CENTER          90

#define PIN_MOTOR_R_PWM  5
#define PIN_MOTOR_R_DIR  4
#define PIN_MOTOR_L_PWM  6
#define PIN_MOTOR_L_DIR  7

#define PIN_ENCODER_1    2
#define PIN_ENCODER_2    3

#define STOP_DISTANCE    7
#define SLOW_DISTANCE   20
#define NORMAL_SPEED   128
#define CREEP_SPEED     50
#define RIGHT_ADJUST     0
#define LEFT_ADJUST     15

/*
** ---------------------------------- Hardware Helper Classes -----------------------
*/

typedef enum {
  stateINITIALIZE = 0,
  stateMOVE,
  stateFINISHED
} state_t;

typedef enum {
  eNONE = 0,
  eOBJ_LONG,
  eOBJ_SHORT,
  eOBJ_BLOCK
} event_t;

typedef void (*action_t)();

typedef struct {
  state_t  nextState;
  action_t actionToDo;
} stateAction_t;

stateAction_t stateMatrix[3][4] = {
  { {stateMOVE, actionInitialize}, {stateMOVE, actionInitialize}, {stateMOVE, actionInitialize},    {stateFINISHED, actionStop} },
  { {stateMOVE, actionNone},       {stateMOVE, actionMoveForward}, {stateMOVE, actionCreepForward},  {stateFINISHED, actionStop} },
  { {stateFINISHED, actionBlink},  {stateFINISHED, actionBlink},   {stateFINISHED, actionBlink},     {stateFINISHED, actionBlink} }
};

Servo   pingServo;
Ping    sensor(PIN_PING);
Encoder encoderA(PIN_ENCODER_1);
Encoder encoderB(PIN_ENCODER_2);
Motor   motors(PIN_MOTOR_R_DIR, PIN_MOTOR_R_PWM, PIN_MOTOR_L_DIR, PIN_MOTOR_L_PWM);
state_t currentState;
event_t currentEvent;
uint8_t readSensor;

void setup()
{
  // Visual indicator that we are finished with our task
  pinMode(PIN_LED, OUTPUT);
  
  // Encoder routines wrap methods in our Encoder class
  attachInterrupt(0, encWrapperA, CHANGE);
  attachInterrupt(1, encWrapperB, CHANGE);

  // Attach to our servo and place it looking forward
  pingServo.attach(PIN_PING_SERVO);
  pingServo.write(CENTER);
  delay(100);
  
  // Set our beginning motor adjustments
  motors.setRightAdjust(RIGHT_ADJUST);
  motors.setLeftAdjust(LEFT_ADJUST);
  
  // Set our initial state and event to run any state machine
  // setup code.
  currentState = stateINITIALIZE;
  currentEvent = eNONE;
  readSensor   = 0;

  // initialize Timer2 for our Event/Sensor code
  noInterrupts();
  TCCR2A=0;
  TCCR2B=0;
  TCNT2 =0;

  OCR2A=31250;              // compare match register 8MHz/256/1000Hz
  TCCR2B |= (1 << WGM22);   // CTC mode
  TCCR2B |= (1 << CS22);    // 256 prescaler
  TIMSK2 |= (1 << OCIE2A);  // enable timer compare interrupt
  interrupts();
  
  DebugCode(  
    Serial.begin(9600);
    Serial.println("Start");
  )
}

/*
** Set our sensor indicator that it's time to read a distance
*/
ISR(TIMER2_COMPA_vect)
{
  readSensor = 1;
}

void loop() {
  if (readSensor)
  {
    getSensorReading();
    readSensor = 0;
  }
  
  stateEval();
}

/*
** Evaluate our current state and the current event to see what
** we need to do next.
*/
void stateEval()
{
  stateAction_t sa = stateMatrix[currentState][currentEvent];
  currentState     = sa.nextState;
  
  (*sa.actionToDo)();
}

/*
** Perform a sensor reading based on our timer.  We need an average
** of 3 sensor readings to ensure we are not getting a false zero
** in our readings.  Then set our event as either a long distance,
** short distance or that we are about to crash.
*/
void getSensorReading()
{
  long total    = 0;
  long distance = 0;

  for (uint8_t x = 0; x < 3; x++) {
    sensor.pulse();
    total += sensor.getCentimeters();
    delay(10);
  }
  
  distance = total / 3;
  
  if (distance > SLOW_DISTANCE) {
    DebugCode(
      Serial.println("eOBJ_LONG event");
    )
    
    currentEvent = eOBJ_LONG;
  } else if (distance > STOP_DISTANCE) {
    DebugCode(
      Serial.println("eOBJ_SHORT event");
    )
    
    currentEvent = eOBJ_SHORT;
  } else {
    DebugCode(
      Serial.println("eOBJ_BLOCK event");
    )
    
    currentEvent = eOBJ_BLOCK;
  }
  
  DebugCode(
    Serial.print("readSensor cm=");
    Serial.println(distance);
  )
}

//-------------------Actions-------------------------

/*
** Do nothing action to help fill in our state machine with a
** function pointer instead of a null value that will mess up
** our function pointer call in the stateEval routine.
*/
void actionNone()
{
  DebugCode(
    Serial.println("actionNone");
  )
}

/*
** General initialization code for the robot
*/
void actionInitialize()
{
  DebugCode(
    Serial.println("actionInitialize");
  )
}

/*
** Move forward at our normal speed since we are not close to any
** objects.
*/
void actionMoveForward()
{
  DebugCode(
    Serial.println("actionMoveForward");
  )
  
  motors.drive(NORMAL_SPEED, Motor::FORWARD);
  displayEncoders();
}

/*
** Creep forward at our minimum speed since we are close to an
** object.
*/
void actionCreepForward()
{
  DebugCode(
    Serial.println("actionCreepForward");
  )
  
  motors.drive(CREEP_SPEED, Motor::FORWARD);
  displayEncoders();
}

/*
** Stop all forward motion.
*/
void actionStop()
{
  DebugCode(
    Serial.println("actionStop");
  )
  
  motors.stop();
  displayEncoders();
}

/*
** Blink our LED to show that we are finished with
** our task.
*/
void actionBlink()
{
  DebugCode(
    Serial.println("actionBlink");
  )
  
  digitalWrite(PIN_LED, digitalRead(PIN_LED) ^ 1);
  delay(500);
}

//-------------------Encoder Test--------------------

void encWrapperA() {
  encoderA.isrUpdate();
}

void encWrapperB() {
  encoderB.isrUpdate();
}

void displayEncoders() {
  long a = encoderA.getCount();
  long b = encoderB.getCount();
  
  Serial.print("Encoder 1: ");
  Serial.print(a);
  Serial.print(" Encoder 2: ");
  Serial.print(b);
  Serial.print(" Diff: ");
  Serial.print(a - b);
  Serial.println();
}
